"""
Profiling hooks

This module contains a couple of decorators (`profile` and `coverage`) that
can be used to wrap functions and/or methods to produce profiles and line
coverage reports.

Usage example (Python 2.4 or newer):

    from profilehooks import profile, coverage

    @profile    # or @coverage
    def fn(n):
        if n < 2: return 1
        else: return n * fn(n-1)

    print fn(42)

Usage example (Python 2.3 or older):

    from profilehooks import profile, coverage

    def fn(n):
        if n < 2: return 1
        else: return n * fn(n-1)

    # Now wrap that function in a decorator
    fn = profile(fn) # or coverage(fn)

    print fn(42)

Reports for all thusly decorated functions will be printed to sys.stdout
on program termination.  You can alternatively request for immediate
reports for each call by passing immediate=True to the profile decorator.

There's also a @timecall decorator for printing the time to sys.stderr
every time a function is called, when you just want to get a rough measure
instead of a detailed (but costly) profile.

Caveats

  A thread on python-dev convinced me that hotshot produces bogus numbers.
  See http://mail.python.org/pipermail/python-dev/2005-November/058264.html

  I don't know what will happen if a decorated function will try to call
  another decorated function.  All decorators probably need to explicitly
  support nested profiling (currently TraceFuncCoverage is the only one that
  supports this, while HotShotFuncProfile has support for recursive functions.)

  Profiling with hotshot creates temporary files (*.prof and *.prof.pickle for
  profiling, *.cprof for coverage) in the current directory.  These files are
  not cleaned up.  (In fact, the *.pickle versions are intentionally written
  out in case you want to look at the profiler results without having to parse
  the big *.prof file with hotshot.stats.load, which takes ages.  Just unpickle
  the file and you'll get a pstats object.)

  Coverage analysis with hotshot seems to miss some executions resulting in
  lower line counts and some lines errorneously marked as never executed.  For
  this reason coverage analysis now uses trace.py which is slower, but more
  accurate.

  Decorating functions causes doctest.testmod() to ignore doctests in those
  functions.

Copyright (c) 2004--2006 Marius Gedminas <marius@pov.lt>

Released under the MIT licence since December 2006:

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

(Previously it was distributed under the GNU General Public Licence.)
"""
# $Id: profilehooks.py,v 1.2 2008/07/18 14:51:08 tflorac Exp $

__author__ = "Marius Gedminas (marius@gedmin.as)"
__copyright__ = "Copyright 2004-2006, Marius Gedminas"
__license__ = "MIT"
__version__ = "1.0"
__date__ = "2006-12-06"


import atexit
import inspect
import sys
import re
import cPickle as pickle

# For profiling
from profile import Profile
import pstats

# For hotshot profiling (inaccurate!)
import hotshot
import hotshot.stats

# For trace.py coverage
import trace

# For hotshot coverage (inaccurate!; uses undocumented APIs; might break)
import _hotshot
import hotshot.log

# For timecall
import time


def profile(fn=None, skip=0, filename=None, immediate=False):
    """Mark `fn` for profiling.

    If `immediate` is False, profiling results will be printed to sys.stdout on
    program termination.  Otherwise results will be printed after each call.

    If `skip` is > 0, first `skip` calls to `fn` will not be profiled.

    If `filename` is specified, the profile stats will be pickled and stored in
    a file.

    Usage:

        def fn(...):
            ...
        fn = profile(fn, skip=1)

    If you are using Python 2.4, you should be able to use the decorator
    syntax:

        @profile(skip=3)
        def fn(...):
            ...

    or just

        @profile
        def fn(...):
            ...

    """
    if fn is None: # @profile() syntax -- we are a decorator maker
        def decorator(fn):
            return profile(fn, skip=skip, filename=filename,
                           immediate=immediate)
        return decorator
    # @profile syntax -- we are a decorator.
    fp = FuncProfile(fn, skip=skip, filename=filename, immediate=immediate)
    # or HotShotFuncProfile
    # We cannot return fp or fp.__call__ directly as that would break method
    # definitions, instead we need to return a plain function.
    def new_fn(*args, **kw):
        return fp(*args, **kw)
    new_fn.__doc__ = fn.__doc__
    return new_fn


def coverage(fn):
    """Mark `fn` for line coverage analysis.

    Results will be printed to sys.stdout on program termination.

    Usage:

        def fn(...):
            ...
        fn = coverage(fn)

    If you are using Python 2.4, you should be able to use the decorator
    syntax:

        @coverage
        def fn(...):
            ...

    """
    fp = TraceFuncCoverage(fn) # or HotShotFuncCoverage
    # We cannot return fp or fp.__call__ directly as that would break method
    # definitions, instead we need to return a plain function.
    def new_fn(*args, **kw):
        return fp(*args, **kw)
    new_fn.__doc__ = fn.__doc__
    return new_fn


def coverage_with_hotshot(fn):
    """Mark `fn` for line coverage analysis.

    Uses the 'hotshot' module for fast coverage analysis.

    BUG: Produces inaccurate results.

    See the docstring of `coverage` for usage examples.
    """
    fp = HotShotFuncCoverage(fn)
    # We cannot return fp or fp.__call__ directly as that would break method
    # definitions, instead we need to return a plain function.
    def new_fn(*args, **kw):
        return fp(*args, **kw)
    new_fn.__doc__ = fn.__doc__
    return new_fn


class FuncProfile:
    """Profiler for a function (uses profile)."""

    # This flag is shared between all instances
    in_profiler = False

    def __init__(self, fn, skip=0, filename=None, immediate=False):
        """Creates a profiler for a function.

        Every profiler has its own log file (the name of which is derived from
        the function name).

        HotShotFuncProfile registers an atexit handler that prints profiling
        information to sys.stderr when the program terminates.

        The log file is not removed and remains there to clutter the current
        working directory.
        """
        self.fn = fn
        self.filename = filename
        self.immediate = immediate
        self.stats = pstats.Stats(Profile())
        self.ncalls = 0
        self.skip = skip
        self.skipped = 0
        atexit.register(self.atexit)

    def __call__(self, *args, **kw):
        """Profile a singe call to the function."""
        self.ncalls += 1
        if self.skip > 0:
            self.skip -= 1
            self.skipped += 1
            return self.fn(*args, **kw)
        if FuncProfile.in_profiler:
            # handle recursive calls
            return self.fn(*args, **kw)
        # You cannot reuse the same profiler for many calls and accumulate
        # stats that way.  :-/
        profiler = Profile()
        try:
            FuncProfile.in_profiler = True
            return profiler.runcall(self.fn, *args, **kw)
        finally:
            FuncProfile.in_profiler = False
            self.stats.add(profiler)
            if self.immediate:
                self.print_stats()
                self.reset_stats()

    def print_stats(self):
        """Print profile information to sys.stdout."""
        funcname = self.fn.__name__
        filename = self.fn.func_code.co_filename
        lineno = self.fn.func_code.co_firstlineno
        print
        print "*** PROFILER RESULTS ***"
        print "%s (%s:%s)" % (funcname, filename, lineno)
        print "function called %d times" % self.ncalls,
        if self.skipped:
            print "(%d calls not profiled)" % self.skipped
        else:
            print
        print
        stats = self.stats
        if self.filename:
            pickle.dump(stats, file(self.filename, 'w'))
        stats.strip_dirs()
        stats.sort_stats('cumulative', 'time', 'calls')
        stats.print_stats(40)

    def reset_stats(self):
        """Reset accumulated profiler statistics."""
        self.stats = pstats.Stats(Profile())
        self.ncalls = 0
        self.skipped = 0

    def atexit(self):
        """Stop profiling and print profile information to sys.stdout.

        This function is registered as an atexit hook.
        """
        if not self.immediate:
            self.print_stats()


class HotShotFuncProfile:
    """Profiler for a function (uses hotshot)."""

    # This flag is shared between all instances
    in_profiler = False

    def __init__(self, fn, skip=0, filename=None):
        """Creates a profiler for a function.

        Every profiler has its own log file (the name of which is derived from
        the function name).

        HotShotFuncProfile registers an atexit handler that prints profiling
        information to sys.stderr when the program terminates.

        The log file is not removed and remains there to clutter the current
        working directory.
        """
        self.fn = fn
        self.filename = filename
        if self.filename:
            self.logfilename = filename + ".raw"
        else:
            self.logfilename = fn.__name__ + ".prof"
        self.profiler = hotshot.Profile(self.logfilename)
        self.ncalls = 0
        self.skip = skip
        self.skipped = 0
        atexit.register(self.atexit)

    def __call__(self, *args, **kw):
        """Profile a singe call to the function."""
        self.ncalls += 1
        if self.skip > 0:
            self.skip -= 1
            self.skipped += 1
            return self.fn(*args, **kw)
        if HotShotFuncProfile.in_profiler:
            # handle recursive calls
            return self.fn(*args, **kw)
        try:
            HotShotFuncProfile.in_profiler = True
            return self.profiler.runcall(self.fn, *args, **kw)
        finally:
            HotShotFuncProfile.in_profiler = False

    def atexit(self):
        """Stop profiling and print profile information to sys.stderr.

        This function is registered as an atexit hook.
        """
        self.profiler.close()
        funcname = self.fn.__name__
        filename = self.fn.func_code.co_filename
        lineno = self.fn.func_code.co_firstlineno
        print
        print "*** PROFILER RESULTS ***"
        print "%s (%s:%s)" % (funcname, filename, lineno)
        print "function called %d times" % self.ncalls,
        if self.skipped:
            print "(%d calls not profiled)" % self.skipped
        else:
            print
        print
        stats = hotshot.stats.load(self.logfilename)
        # hotshot.stats.load takes ages, and the .prof file eats megabytes, but
        # a pickled stats object is small and fast
        if self.filename:
            pickle.dump(stats, file(self.filename, 'w'))
        # it is best to pickle before strip_dirs
        stats.strip_dirs()
        stats.sort_stats('cumulative', 'time', 'calls')
        stats.print_stats(200)


class HotShotFuncCoverage:
    """Coverage analysis for a function (uses _hotshot).

    HotShot coverage is reportedly faster than trace.py, but it appears to
    have problems with exceptions; also line counts in coverage reports
    are generally lower from line counts produced by TraceFuncCoverage.
    Is this my bug, or is it a problem with _hotshot?
    """

    def __init__(self, fn):
        """Creates a profiler for a function.

        Every profiler has its own log file (the name of which is derived from
        the function name).

        HotShotFuncCoverage registers an atexit handler that prints profiling
        information to sys.stderr when the program terminates.

        The log file is not removed and remains there to clutter the current
        working directory.
        """
        self.fn = fn
        self.logfilename = fn.__name__ + ".cprof"
        self.profiler = _hotshot.coverage(self.logfilename)
        self.ncalls = 0
        atexit.register(self.atexit)

    def __call__(self, *args, **kw):
        """Profile a singe call to the function."""
        self.ncalls += 1
        return self.profiler.runcall(self.fn, args, kw)

    def atexit(self):
        """Stop profiling and print profile information to sys.stderr.

        This function is registered as an atexit hook.
        """
        self.profiler.close()
        funcname = self.fn.__name__
        filename = self.fn.func_code.co_filename
        lineno = self.fn.func_code.co_firstlineno
        print
        print "*** COVERAGE RESULTS ***"
        print "%s (%s:%s)" % (funcname, filename, lineno)
        print "function called %d times" % self.ncalls
        print
        fs = FuncSource(self.fn)
        reader = hotshot.log.LogReader(self.logfilename)
        for what, (filename, lineno, funcname), tdelta in reader:
            if filename != fs.filename:
                continue
            if what == hotshot.log.LINE:
                fs.mark(lineno)
            if what == hotshot.log.ENTER:
                # hotshot gives us the line number of the function definition
                # and never gives us a LINE event for the first statement in
                # a function, so if we didn't perform this mapping, the first
                # statement would be marked as never executed
                if lineno == fs.firstlineno:
                    lineno = fs.firstcodelineno
                fs.mark(lineno)
        reader.close()
        print fs


class TraceFuncCoverage:
    """Coverage analysis for a function (uses trace module).

    HotShot coverage analysis is reportedly faster, but it appears to have
    problems with exceptions.
    """

    # Shared between all instances so that nested calls work
    tracer = trace.Trace(count=True, trace=False,
                         ignoredirs=[sys.prefix, sys.exec_prefix])

    # This flag is also shared between all instances
    tracing = False

    def __init__(self, fn):
        """Creates a profiler for a function.

        Every profiler has its own log file (the name of which is derived from
        the function name).

        TraceFuncCoverage registers an atexit handler that prints profiling
        information to sys.stderr when the program terminates.

        The log file is not removed and remains there to clutter the current
        working directory.
        """
        self.fn = fn
        self.logfilename = fn.__name__ + ".cprof"
        self.ncalls = 0
        atexit.register(self.atexit)

    def __call__(self, *args, **kw):
        """Profile a singe call to the function."""
        self.ncalls += 1
        if TraceFuncCoverage.tracing:
            return self.fn(*args, **kw)
        try:
            TraceFuncCoverage.tracing = True
            return self.tracer.runfunc(self.fn, *args, **kw)
        finally:
            TraceFuncCoverage.tracing = False

    def atexit(self):
        """Stop profiling and print profile information to sys.stderr.

        This function is registered as an atexit hook.
        """
        funcname = self.fn.__name__
        filename = self.fn.func_code.co_filename
        lineno = self.fn.func_code.co_firstlineno
        print
        print "*** COVERAGE RESULTS ***"
        print "%s (%s:%s)" % (funcname, filename, lineno)
        print "function called %d times" % self.ncalls
        print
        fs = FuncSource(self.fn)
        for (filename, lineno), count in self.tracer.counts.items():
            if filename != fs.filename:
                continue
            fs.mark(lineno, count)
        print fs
        never_executed = fs.count_never_executed()
        if never_executed:
            print "%d lines were not executed." % never_executed


class FuncSource:
    """Source code annotator for a function."""

    blank_rx = re.compile(r"^\s*finally:\s*(#.*)?$")

    def __init__(self, fn):
        self.fn = fn
        self.filename = inspect.getsourcefile(fn)
        self.source, self.firstlineno = inspect.getsourcelines(fn)
        self.sourcelines = {}
        self.firstcodelineno = self.firstlineno
        self.find_source_lines()

    def find_source_lines(self):
        """Mark all executable source lines in fn as executed 0 times."""
        strs = trace.find_strings(self.filename)
        lines = trace.find_lines_from_code(self.fn.func_code, strs)
        self.firstcodelineno = sys.maxint
        for lineno in lines:
            self.firstcodelineno = min(self.firstcodelineno, lineno)
            self.sourcelines.setdefault(lineno, 0)
        if self.firstcodelineno == sys.maxint:
            self.firstcodelineno = self.firstlineno

    def mark(self, lineno, count=1):
        """Mark a given source line as executed count times.

        Multiple calls to mark for the same lineno add up.
        """
        self.sourcelines[lineno] = self.sourcelines.get(lineno, 0) + count

    def count_never_executed(self):
        """Count statements that were never executed."""
        lineno = self.firstlineno
        counter = 0
        for line in self.source:
            if self.sourcelines.get(lineno) == 0:
                if not self.blank_rx.match(line):
                    counter += 1
            lineno += 1
        return counter

    def __str__(self):
        """Return annotated source code for the function."""
        lines = []
        lineno = self.firstlineno
        for line in self.source:
            counter = self.sourcelines.get(lineno)
            if counter is None:
                prefix = ' ' * 7
            elif counter == 0:
                if self.blank_rx.match(line):
                    prefix = ' ' * 7
                else:
                    prefix = '>' * 6 + ' '
            else:
                prefix = '%5d: ' % counter
            lines.append(prefix + line)
            lineno += 1
        return ''.join(lines)


def timecall(fn):
    """Wrap `fn` and print its execution time.

    Example:

        @timecall
        def somefunc(x, y):
            time.sleep(x * y)

        somefunc(2, 3)

    will print

        somefunc: 6.0 seconds

    """
    def new_fn(*args, **kw):
        try:
            start = time.time()
            return fn(*args, **kw)
        finally:
            duration = time.time() - start
            funcname = fn.__name__
            filename = fn.func_code.co_filename
            lineno = fn.func_code.co_firstlineno
            print >> sys.stderr, "\n  %s (%s:%s):\n    %.3f seconds\n" % (
                                        funcname, filename, lineno, duration)
    new_fn.__doc__ = fn.__doc__
    return new_fn




### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages
import pytz

# import Zope3 interfaces
from zope.schema.interfaces import IChoice, IVocabularyFactory

# import local interfaces

# import Zope3 packages
from zope.interface import implements, classProvides
from zope.schema import Choice
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

# import local packages


class TimezonesVocabulary(SimpleVocabulary):

    classProvides(IVocabularyFactory)

    def __init__(self, *args, **kw):
        terms = [SimpleTerm(t, t, t) for t in pytz.all_timezones]
        super(TimezonesVocabulary, self).__init__(terms)


class ITimezone(IChoice):
    """Marker interface for timezone field"""


class Timezone(Choice):
    """Timezone choice field"""

    implements(ITimezone)

    def __init__(self, **kw):
        if 'vocabulary' in kw:
            kw.pop('vocabulary')
        if 'default' not in kw:
            kw['default'] = u'GMT'
        super(Timezone, self).__init__(vocabulary='ZTFY timezones', **kw)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages
import pytz

# import Zope3 interfaces
from zope.interface.common.idatetime import ITZInfo
from zope.publisher.interfaces.browser import IBrowserRequest

# import local interfaces
from ztfy.utils.timezone.interfaces import IServerTimezone

# import Zope3 packages
from zope.component import adapter, queryUtility
from zope.interface import implementer

# import local packages


GMT = pytz.timezone('GMT')
_tz = pytz.timezone('Europe/Paris')
tz = _tz

@implementer(ITZInfo)
@adapter(IBrowserRequest)
def tzinfo(request=None):
    util = queryUtility(IServerTimezone)
    if util is not None:
        return pytz.timezone(util.timezone)
    return GMT


def tztime(value):
    if not value:
        return None
    if not value.tzinfo:
        value = GMT.localize(value)
    return value.astimezone(tzinfo())


def gmtime(value):
    if not value:
        return None
    if not value.tzinfo:
        value = GMT.localize(value)
    return value.astimezone(GMT)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages
from zope.interface import Interface

# import local packages


def getParent(context, interface=Interface, allow_context=True):
    """Get first parent of the given context that implements given interface"""
    if allow_context:
        parent = context
    else:
        parent = getattr(context, '__parent__', None)
    while parent is not None:
        if interface.providedBy(parent):
            return interface(parent)
        parent = getattr(parent, '__parent__', None)
    return None


# copied from fanstatic (which copied it from zope.dottedname !)
def resolve(name, module=None):
    name = name.split('.')
    if not name[0]:
        if module is None:
            raise ValueError("relative name without base module")
        module = module.split('.')
        name.pop(0)
        while not name[0]:
            module.pop()
            name.pop(0)
        name = module + name

    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used += '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)

    return found



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.tales.interfaces import ITALESFunctionNamespace

# import local interfaces
from ztfy.utils.tal.interfaces import IHTMLTalesAPI

# import Zope3 packages
from zope.interface import implements

# import local packages
from ztfy.utils.html import htmlToText
from ztfy.utils.text import textToHTML, Renderer


class HTMLTalesAdapter(object):

    implements(IHTMLTalesAPI, ITALESFunctionNamespace)

    def __init__(self, context):
        self.context = context

    def setEngine(self, engine):
        self.request = engine.vars['request']

    def renderer(self):
        return Renderer(self.context)

    def totext(self):
        if not self.context:
            return u''
        return htmlToText(self.context)

    def text(self):
        if not self.context:
            return u''
        return textToHTML(self.context, 'zope.source.plaintext')

    def stx(self):
        if not self.context:
            return u''
        return textToHTML(self.context, 'zope.source.stx')

    def rest(self):
        if not self.context:
            return u''
        return textToHTML(self.context, 'zope.source.rest')



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages
import base64
import cookielib
import urllib2
import xmlrpclib

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages

# import local packages


class XMLRPCCookieAuthTransport(xmlrpclib.Transport):
    """An XML-RPC transport handling authentication via cookies"""

    def __init__(self, user_agent, credentials=(), cookies=None):
        xmlrpclib.Transport.__init__(self)
        self.user_agent = user_agent
        self.credentials = credentials
        self.cookies = cookies

    # override the send_host hook to also send authentication info
    def send_host(self, connection, host):
        xmlrpclib.Transport.send_host(self, connection, host)
        if (self.cookies is not None) and (len(self.cookies) > 0):
            for cookie in self.cookies:
                connection.putheader('Cookie', '%s=%s' % (cookie.name, cookie.value))
        elif self.credentials:
            auth = 'Basic %s' % base64.encodestring("%s:%s" % self.credentials).strip()
            connection.putheader('Authorization', auth)

    def request(self, host, handler, request_body, verbose=False):

        # dummy request class for extracting cookies
        class CookieRequest(urllib2.Request):
            pass

        # dummy response class for extracting cookies
        class CookieResponse:
            def __init__(self, headers):
                self.headers = headers
            def info(self):
                return self.headers

        # issue XML-RPC request
        connection = self.make_connection(host)
        self.verbose = verbose
        if verbose:
            connection.set_debuglevel(1)
        self.send_request(connection, handler, request_body)
        self.send_host(connection, host)
        self.send_user_agent(connection)
        # get response
        self.send_content(connection, request_body)
        errcode, errmsg, headers = connection.getreply()
        # extract cookies from response headers
        crequest = CookieRequest('http://%s/' % host)
        cresponse = CookieResponse(headers)
        if self.cookies is not None:
            self.cookies.extract_cookies(cresponse, crequest)
        if errcode != 200:
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)
        try:
            sock = connection._conn.sock
        except AttributeError:
            sock = None
        return self._parse_response(connection.getfile(), sock)


def getClient(uri, credentials=(), verbose=False):
    """Get an XML-RPC client which supports basic authentication"""
    transport = XMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY basic implementation)', credentials)
    return xmlrpclib.Server(uri, transport=transport, verbose=verbose)


def getClientWithCookies(uri, credentials=(), verbose=False):
    """Get an XML-RPC client which supports authentication throught cookies"""
    transport = XMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY cookie implementation)', credentials, cookielib.CookieJar())
    return xmlrpclib.Server(uri, transport=transport, verbose=verbose)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages
from zope.component import createObject, queryMultiAdapter

# import local packages
from ztfy.utils.request import getRequest


def textStart(text, length, max=0):
    """Get first words of given text with maximum given length
    
    If @max is specified, text is shortened only if remaining text is longer than @max
    
    @param text: initial text
    @param length: maximum length of resulting text
    @param max: if > 0, @text is shortened only if remaining text is longer than max
    """
    result = text or ''
    if length > len(result):
        return result
    index = length - 1
    text_length = len(result)
    while (index > 0) and (result[index] != ' '):
        index -= 1
    if (index > 0) and (text_length > index + max):
        return result[:index] + '&#133;'
    return text


def textToHTML(text, renderer='zope.source.plaintext', request=None):
    if request is None:
        request = getRequest()
    formatter = createObject(renderer, text)
    renderer = queryMultiAdapter((formatter, request), name=u'')
    return renderer.render()


class Renderer(object):

    def __init__(self, context):
        self.context = context

    def render(self, renderer, request=None):
        if not self.context:
            return u''
        return textToHTML(self.context, renderer, request)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages

# import Zope3 interfaces
from z3c.form.interfaces import IFileWidget
from zope.schema.interfaces import IBytes

# import local interfaces

# import Zope3 packages
from z3c.form.converter import FileUploadDataConverter as BaseDataConverter
from zope.component import adapts

# import local packages


class FileUploadDataConverter(BaseDataConverter):

    adapts(IBytes, IFileWidget)

    def toWidgetValue(self, value):
        return value or ''



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.annotation.interfaces import IAnnotations
from zope.app.file.interfaces import IFile
from zope.catalog.interfaces import ICatalog
from zope.container.interfaces import IContainer
from zope.intid.interfaces import IIntIds

# import local interfaces

# import Zope3 packages
from zope.component import queryUtility, getAllUtilitiesRegisteredFor

# import local packages
from ztfy.utils import request as request_utils

from ztfy.utils import _


#
# IntIds utility functions
#

def getIntIdUtility(name='', request=None, context=None):
    """Look for a named IIntIds utility"""
    if request is None:
        request = request_utils.queryRequest()
    intids = request_utils.getRequestData('IntIdsUtility::' + name, request)
    if intids is None:
        intids = queryUtility(IIntIds, name, context=context)
        if (request is not None) and (intids is not None):
            request_utils.setRequestData('IntIdsUtility::' + name, intids, request)
    return intids


def getObjectId(object, intids_name='', request=None, context=None):
    """Look for an object Id as recorded by given IIntIds utility"""
    if object is None:
        return None
    if request is None:
        request = request_utils.queryRequest()
    intids = getIntIdUtility(intids_name, request, context)
    if intids is not None:
        return intids.queryId(object)
    return None


def getObject(id, intids_name='', request=None, context=None):
    """Look for an object recorded by given IIntIds utility and id"""
    if id is None:
        return None
    if request is None:
        request = request_utils.getRequest()
    intids = getIntIdUtility(intids_name, request, context)
    if intids is not None:
        return intids.queryObject(id)
    return None


#
# Catalog utility functions
#

def queryCatalog(name='', context=None):
    """Look for a registered catalog"""
    return queryUtility(ICatalog, name, context=context)


def indexObject(object, catalog_name='', index_name='', request=None, context=None):
    """Index object into a registered catalog"""
    if request is None:
        request = request_utils.getRequest()
    intids = getIntIdUtility('', request, context)
    if intids is not None:
        if ICatalog.providedBy(catalog_name):
            catalog = catalog_name
        else:
            catalog = queryCatalog(catalog_name, context)
        if catalog is not None:
            id = intids.register(object)
            if index_name:
                catalog[index_name].index_doc(id, object)
            else:
                catalog.index_doc(id, object)
            return True
    return False


def unindexObject(object, catalog_name='', index_name='', request=None, context=None):
    """Remove object from a registered catalog"""
    if request is None:
        request = request_utils.getRequest()
    id = getObjectId(object, '', request, context)
    if id is not None:
        if ICatalog.providedBy(catalog_name):
            catalog = catalog_name
        else:
            catalog = queryCatalog(catalog_name, context)
        if catalog is not None:
            if index_name:
                catalog[index_name].unindex_doc(id)
            else:
                catalog.unindex_doc(id)
            return True
    return False


def _indexObject(object, intids, catalogs):
    """Index object data into given set of catalogs"""
    id = intids.register(object)
    for catalog in catalogs:
        catalog.index_doc(id, object)

def _indexObjectValues(object, intids, catalogs):
    """Index object values into given set of catalogs"""
    container = IContainer(object, None)
    if container is not None:
        for subobject in container.values():
            _indexAllObject(subobject, intids, catalogs)

def _indexObjectAnnotations(object, intids, catalogs):
    """Index object annotations into given set of catalogs"""
    annotations = IAnnotations(object, None)
    if annotations is not None:
        keys = annotations.keys()
        for key in keys:
            _indexAllObject(annotations[key], intids, catalogs)
            file = IFile(annotations[key], None)
            if file is not None:
                _indexObject(file, intids, catalogs)

def _indexAllObject(object, intids, catalogs):
    """Index object, object values and annotations into given set of catalogs"""
    _indexObject(object, intids, catalogs)
    _indexObjectValues(object, intids, catalogs)
    _indexObjectAnnotations(object, intids, catalogs)

def indexAllObjectValues(object, context=None):
    """Reindex a whole container properties and contents (including annotations) into site's catalogs"""
    if context is None:
        context = object
    intids = queryUtility(IIntIds, context=context)
    if intids is not None:
        catalogs = getAllUtilitiesRegisteredFor(ICatalog, context)
        if catalogs:
            _indexAllObject(object, intids, catalogs)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.intid.interfaces import IIntIds

# import local interfaces
from ztfy.utils.interfaces import INewSiteManagerEvent

# import Zope3 packages
from zope.component import getUtility
from zope.interface import implements
from zope.location import locate

# import local packages


class NewSiteManagerEvent(object):
    """Event notified when a new site manager is created"""

    implements(INewSiteManagerEvent)

    def __init__(self, obj):
        self.object = obj


def locateAndRegister(contained, parent, key, intids=None):
    locate(contained, parent)
    if intids is None:
        intids = getUtility(IIntIds)
    intids.register(contained)
    parent[key] = contained



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.component.interfaces import IObjectEvent

# import local interfaces

# import Zope3 packages
from zope.interface import Interface

# import local packages


class INewSiteManagerEvent(IObjectEvent):
    """Event interface for new site manager event"""


#
# Generic list interface
#

class IListInfo(Interface):
    """Custom interface used to handle list-like components"""

    def count():
        """Get list items count"""

    def index():
        """Get position of the given item"""

    def __contains__():
        """Check if given value is int list"""

    def __getitem__():
        """Return item at given position"""

    def __iter__():
        """Iterator over list items"""


class IListWriter(Interface):
    """Writer interface for list-like components"""

    def append():
        """Append value to list"""

    def extend():
        """Extend list with given items"""

    def insert():
        """Insert item to given index"""

    def pop():
        """Pop item from list and returns it"""

    def remove():
        """Remove given item from list"""

    def reverse():
        """Sort list in reverse order"""

    def sort():
        """Sort list"""


class IList(IListInfo, IListWriter):
    """Marker interface for list-like components"""


#
# Generic dict interface
#

class IDictInfo(Interface):
    """Custom interface used to handle dict-like components"""

    def keys():
        """Get list of keys for the dict"""

    def has_key(key):
        """Check to know if dict includes the given key"""

    def get(key, default=None):
        """Get given key or default from dict"""

    def copy():
        """Duplicate content of dict"""

    def __contains__(key):
        """Check if given key is in dict"""

    def __getitem__(key):
        """Get given key value from dict"""

    def __iter__():
        """Iterator over dictionnary keys"""


class IDictWriter(Interface):
    """Writer interface for dict-like components"""

    def clear():
        """Clear dict"""

    def update(b):
        """Update dict with given values"""

    def setdefault(key, failobj=None):
        """Create value for given key if missing"""

    def pop(key, *args):
        """Remove and return given key from dict"""

    def popitem():
        """Pop item from dict"""

    def __setitem__(key, value):
        """Update given key with given value"""

    def __delitem__(key):
        """Remove selected key from dict"""


class IDict(IDictInfo, IDictWriter):
    """Marker interface for dict-like components"""



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages
from zope.interface import Interface

# import local packages


class ITextStartTalesAPI(Interface):
    """'start' TALES namespace interface"""

    def __getattr__(attr):
        """Get first characters of adapted text, without cutting words"""


class ITextOutputTalesAPI(Interface):
    """'text' TALES namespace interface"""

    def js():
        """Convert adapted text to a JavaScript compatible output"""

    def noquotes():
        """Remove double quotes from adapted text"""

    def breaks():
        """Replace '|' by newlines in adapted text"""


class IHTMLTalesAPI(Interface):
    """'html' TALES namespace interface"""

    def renderer():
        """Get text to HTML renderer"""

    def totext():
        """Convert adapted HTML content to text"""

    def text():
        """Convert adapted text to HTML"""

    def stx():
        """Convert adapted StructuredText to HTML"""

    def rest():
        """Convert adapted reStructuredText to HTML"""


class IRequestDataTalesAPI(Interface):
    """Request 'data' TALES namespace interface"""

    def __getattr__(attr):
        """Get request data for given attribute"""


class ISessionDataTalesAPI(Interface):
    """Session data TALES namespace interface"""

    def __getattr__(attr):
        """Get session data for given app and key"""
